import cv2
import gimmick
import numpy as np

def load_images():
    # from sklearn import datasets
    # images = datasets.load_digits().images

    import numpy as np
    import tensorflow as tf
    (images, _), (_, _) = tf.keras.datasets.mnist.load_data()
    images = np.array([cv2.resize(x, (32, 32)) for x in images])
    return images


def train_model(modelfile, model_type, latent_dimension, num_encoder_layers, num_decoder_layers, optimizer, metrics,
                loss_function, epochs, batch_size, learning_rate, samples_for_code_statistics, images=None):

    print("Training data shape:", images.shape)

    if images is None:
        print("No Image dataset passes hence using Default images for training")
        images = load_images()

    print("Training started .......")
    """ Train and save model """
    model = gimmick.learn(images, algo=model_type, epochs=epochs, batch_size=batch_size, learning_rate=learning_rate,
                          metrics=[metrics], loss_function=loss_function, code_length=latent_dimension,
                          samples_for_code_statistics=samples_for_code_statistics,
                          num_encoder_layers=num_encoder_layers, num_decoder_layers=num_decoder_layers)
    model.save(modelfile)  # Saving model to be used later
    print("Training completed .......")


def get_model_details(modelfile):
    model = gimmick.load(modelfile) # loading model
    return {
        'code_stats': model.code_stats,
        'code_length': model.code_length
    }

def generate_image(model, code_values, random=False):
    print(model.code_stats)
    codes = [code_values] if not random else None
    images_gen = model.generate(1, batch_size=1, codes=codes) # Generate N random samples/images
    print(images_gen.shape)
    if images_gen[0].shape[-1] == 1:
        return cv2.cvtColor(images_gen[0], cv2.COLOR_GRAY2RGB)
    images_gen[0] = cv2.cvtColor(images_gen[0] , cv2.COLOR_BGR2RGB)
    return images_gen[0] #* 12

def generate_code(model, image):
    code = model.generate_code(np.array([image]))
    return code



import math
import pickle
import numpy as np
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers
from tensorflow.keras.callbacks import ModelCheckpoint
from datetime import datetime
from gimmick import constants
from gimmick.models.autoencoder import AutoEncoder


class Model(AutoEncoder):
    """ This algorithm uses a LSTM based encoder-decoder n/w for learning and generating images.
    https://en.wikipedia.org/wiki/Long_short-term_memory
    """

    def build_model_graph(self, images_shape):
        total_pixels = images_shape[0] * images_shape[1] * images_shape[2]

        num_encoder_layers = int(math.log(images_shape[0], 2)) - 2 if self.num_encoder_layers == 'auto' else int(self.num_encoder_layers)
        num_encoder_layers = max(num_encoder_layers, 3)

        num_decoder_layers = int(math.log(images_shape[0], 2))- 2 if self.num_decoder_layers == 'auto' else int(self.num_decoder_layers)
        num_decoder_layers = max(num_decoder_layers, 3)

        log2_code = int(math.log(self.code_length, 2))
        print("num_enoder_layer:\t", num_encoder_layers)
        print("num_decoder_layers:\t", num_decoder_layers)
        print("log2_code:\t\t", log2_code)

        model = keras.Sequential(name="autoencoder_lstm")
        model.add(layers.InputLayer(input_shape=images_shape))
        model.add(layers.Reshape((images_shape[0], images_shape[1] * images_shape[2])))

        # Encoder Layer
        for i in range(1, num_encoder_layers + 1):
            neurons = 2 ** (num_encoder_layers - i + log2_code + 1) # Encoder layer size will be always greater then code_length by multiple of 2
            return_sequences = True if i < num_decoder_layers else False # In Last layer we do not return sequences
            model.add(layers.LSTM(neurons, activation='relu', return_sequences=return_sequences, name="encoder_layer_" + str(i) ))

        # Code Layer
        model.add(layers.Dense(self.code_length, name="code"))
        model.add(layers.Reshape((2, int(self.code_length/2) )))

        # Decoder Layer
        for i in range(1, num_decoder_layers + 1):
            neurons = 2 ** (i + log2_code)  # Decoder layer size will be always greater then code_length by multiple of 2
            return_sequences = True if i < num_decoder_layers else False # In Last layer we do not return sequences
            model.add(layers.LSTM(neurons, activation="relu", return_sequences=return_sequences, name="decoder_layer_" + str(i) ))

        model.add(layers.Dense(total_pixels, activation="relu", name="final_layer"))
        model.add(layers.Reshape(images_shape))

        optimizer =self.optimizer
        optimizer.learning_rate = self.learning_rate

        model.compile(optimizer=optimizer, loss=self.loss_function, metrics=self.metrics)

        print(model.summary())
        self.model = model

        def _code_generator_model():
            layers_ = [layers.InputLayer(input_shape=images_shape), layers.Reshape((images_shape[0], images_shape[1] * images_shape[2]))]
            encoder_layers = [layer.name if 'encoder_layer' in layer.name else None for layer in model.layers]
            num_encoder_layers = len(list(filter(lambda x: x, encoder_layers))) + 2 # 1 Each for (Flatten, Reshape, Code) layer
            layers_.extend(model.layers[:num_encoder_layers])  # Trim all layers except encoder layers

            model_code_generator = keras.Sequential(layers_)
            model_code_generator.build((None, images_shape[0], images_shape[1], images_shape[2]))

            for layer in model_code_generator.layers:
                if list(filter(lambda x: x in layer.name, ['flatten', 'reshape'])):
                    continue
                assert all([np.array_equal(layer.get_weights()[0], model.get_layer(layer.name).get_weights()[0]),
                            np.array_equal(layer.get_weights()[1], model.get_layer(layer.name).get_weights()[1])]),  "%s weights not same" % layer.name

            print(model_code_generator.summary())
            return model_code_generator

        self.model_code_generator = _code_generator_model()

        def _image_generator_model():
            encoder_layers = [layer.name if 'encoder_layer' in layer.name else None for layer in model.layers]
            num_encoder_layers = len(list(filter(lambda x: x, encoder_layers))) + 2 # 1 Each for (Flatten, Reshape, Code) layer

            # Building model
            model_image_generator = keras.Sequential(model.layers[num_encoder_layers:])
            model_image_generator.build((None, self.code_length))
            print(model_image_generator.summary())
            return model_image_generator

        self.model_image_generator = _image_generator_model()



import os

DEFAULT_LEARNING_RATE = 0.001
DEFAULT_TF_MODELFILE = "model.tf"
DEFAULT_TF_CHECKPOINT_DIR = "model.tf"
DEFAULT_TF_CHECKPOINT = os.path.join(DEFAULT_TF_CHECKPOINT_DIR, "checkpoint")
# DEFAULT_TF_GENERATOR_MODELFILE = "model_generator.tf"
# DEFAULT_TF_DISCRIMINATOR_MODELFILE = "model_discriminator.tf"



import os
import pickle
import shutil
import numpy as np
import tensorflow as tf
from datetime import datetime
from gimmick import constants
from tensorflow.keras.callbacks import ModelCheckpoint
from zipfile import ZipFile

class AutoEncoder():
    """ Autoenocoder - Base class for Autoencoders

        **Parameters**

        learning_rate: int
            value passed by user to distributer.learn function
        optimizer:  object
            value passed by user to distributer.learn function
        optimizer_keys:  string
            value passed by user to distributer.learn function
        loss_function: object
            value passed by user to distributer.learn function
        loss_function_keys: string
            value passed by user to distributer.learn function
        metrics: list of string
            value passed by user to distributer.learn function
        metrics_keys: list of string
            value passed by user to distributer.learn function
        code_length: int
            Default 8, Length of intermediate representation or condense space generated by model. In order to generate a random image sample having dimention equal to code_length must be passed.
        num_encoder_layer: int
            Default 'auto', number of layers to be used in encoder, applicable for autoenocder
        num_decoder_layers: int
            Default 'auto', number of layers to be used in decoder, applicable for autoenocder
    """

    def __init__(self, learning_rate=None, optimizer=None, optimizer_keys=None, loss_function=None, loss_function_keys=None, metrics=None, metrics_keys=None,
                 code_length=8, num_encoder_layers=-1, num_decoder_layers=-1):
        self.learning_rate = learning_rate
        self.optimizer = optimizer
        self.loss_function = loss_function
        self.metrics = metrics
        self.num_encoder_layers = num_encoder_layers
        self.num_decoder_layers = num_decoder_layers
        self.code_length = code_length

        self.loss_function_keys = loss_function_keys
        self.optimizer_keys = optimizer_keys
        self.metrics_keys = metrics_keys

    def build_model_graph(self, images_shape):
        """
        Function to build model by learning image representation

        **Parameters**

        images_shape :  array
            shape of the images which are passed as training samples, list of int Eg. [44, 44, 3], this shape denotes that each image is of 44x44x3 diminesion.
        """
        pass


    def train(self, images_train, images_test, epochs=10, batch_size=16, validation_split=0.2):
        """ This function is used to train model on given training images, based on data enoder part learns to compress images into intermediate dimension containing N points, and decoder learns to decipher the intermediate dimension. The indermediate dimention of N points (code_length), when randomly generated will produce new images.

        **Parameters**

        images_train: list
            3D shape of the image, Eg, 128x128x3, 64x64x3 to be used for training model
        images_test: list
            3D shape of the image, Eg, 128x128x3, 64x64x3 to be used for testing model
        epochs: int
            Default 10, number of epochs for model training.
        batch_size: int
            Default 16, batch_size used for training model.
        validation_split: float
            Default 0.2, validation ratio to be used for validating model, it is formed by sampling training data
        """

        startime = datetime.now()
        self.image_shape = images_train[0].shape

        checkpoint = ModelCheckpoint(constants.DEFAULT_TF_MODELFILE, verbose=0, monitor='val_loss', save_best_only=True, mode='auto')
        early_stopping = tf.keras.callbacks.EarlyStopping(monitor='val_loss', patience=5, restore_best_weights=True)

        print("================================= Training ===================================")
        model = self.model
        model.fit(images_train, images_train, batch_size=batch_size, epochs=epochs, validation_split=validation_split,
                  callbacks=[checkpoint, early_stopping], shuffle=True)

        model.save(constants.DEFAULT_TF_MODELFILE) # Save Best model to disk
        print("Total Training time:", datetime.now() - startime)

        print("================================= Evaluating ===================================")
        model.evaluate(images_test, images_test, batch_size=batch_size, verbose=True)

    def generate_code(self, images, batch_size=8, print_model=False):
        image_shape = images[0].shape
        if image_shape[0] != self.image_shape[0] and image_shape[1] != self.image_shape[1] :
            raise Exception("Please pass images with same dimention as passed during training", str(self.image_shape))

        model_code_generator = self.model_code_generator
        if print_model:
            print(model_code_generator.summary())
        codes = model_code_generator.predict(images, batch_size=batch_size, verbose=False)
        return codes

    def prepare_code_statistics(self, images, batch_size=8, sample_size=64, print_model=False):
        """ This function return the statistics for intermedite highly condence space of N Dimention
        which can be used to generate similar samples

        **Parameters**

        images: list
            3D shape of the image, Eg, 128x128x3, 64x64x3
        batch_size: int
            Default 32, batch_size for generating code statistics from learned model, by passing training images as batches of batch_size.
        sample_size: int
            Default 64, number of samples from training images to be used for generating code statistics
        print_model: Boolean
            Default False, whether to print model information
        """
        print("================================= generating code statistics ===================================")

        print("Total samples used to generate code statistics:", sample_size)
        image_shape = images[0].shape
        if image_shape[0] != self.image_shape[0] and image_shape[1] != self.image_shape[1] :
            raise Exception("Please pass images with same dimention as passed during training", str(self.image_shape))

        model_code_generator = self.model_code_generator
        if print_model:
            print(model_code_generator.summary())

        codes = model_code_generator.predict(images[:sample_size], batch_size=batch_size, verbose=False)
        print("codes shape:", codes.shape)

        assert codes.shape[1] == self.code_length, "code_length_passed (%d) and code_length_generated (%d) does not match" % (self.code_length, codes.shape[1])

        print('code 0 =>', codes[0].tolist())
        print('code 1 =>', codes[1].tolist())
        print('code 2 =>', codes[2].tolist())

        code_stats = {
            # "min" : np.min(codes, axis=0),
            # "max" : np.max(codes, axis=0),
            "mean": np.mean(codes, axis=0),
            "std": np.std(codes, axis=0)
        }
        self.code_stats = code_stats
        print("code_stats:", code_stats)
        return codes

    def generate(self, n, codes=None, batch_size=8, print_model=False):
        """ This function is used to train model on given training images, based on data enoder part learns to compress images into intermediate dimension containing N points, and decoder learns to decipher the intermediate dimension. The indermediate dimention of N points (code_length), when randomly generated will produce new images.

        **Parameters**

        n: int
            number of images to be generated by our trained model
        codes: list
            Default: None, if passes then model will use this set of codes for generating images, else it will generate random distribution based on earlier calculated code statistics
        batch_size: int
            Default 16, batch_size used for training model.
        print_model: boolean
            Default False,  whether to print model information
        """
        print("================================= generating samples ===================================")
        code_stats = self.code_stats

        # print(code_stats)
        # Building model

        model_image_generator = self.model_image_generator
        if print_model:
            print(model_image_generator.summary())

        if codes is not None:
            inputs  = codes[:n]
        else:
            # inputs = np.random.normal(code_stats['mean'], code_stats['std'], (n, self.code_length))  # Random samples
            inputs = []
            for i in range(self.code_length):
                inputs.append(np.random.normal(code_stats['mean'][i], code_stats['std'][i], (n,1)))
            inputs = np.concatenate(inputs, axis=1)

        images_generated = model_image_generator.predict(inputs, batch_size=batch_size, verbose=False).astype(np.uint8)
        images_generated[images_generated > 255] = 255
        images_generated[images_generated < 0] = 0
        return images_generated

    def reproduce(self, images, batch_size=8):
        """ Testing function to check whether our model can reprodue exact same image as given as input.

        **Parameters**

        images: list
            N 3D images, Eg, 512x128x128x3, 1024x64x64x3
        batch_size: int
            Default 8, batch_size used for evaluation
        """
        images_generated = self.model.predict(images, batch_size=batch_size, verbose=False).astype(np.uint8)
        images_generated = images_generated.reshape(-1, 8, 8)
        return images_generated

    def save(self, modelfile):
        """ Function save model to a persistat storage

        **Parameters**

        modelfile: string
            Path where model needs to be saved
        """
        if not modelfile.endswith('.zip'):
            raise Exception('modelfile must ends with .zip as extention')

        if os.path.exists(modelfile):
            os.remove(modelfile)

        modelfile_cls = "tf_" + modelfile.split('.')[0] + ".pkl"
        modelfile_tf = "tf_" + modelfile.split('.')[0] + ".h5"
        modelfile_ig_tf = "tf_" + modelfile.split('.')[0] + "_ig.h5"
        modelfile_cg_tf = "tf_" + modelfile.split('.')[0] + "_cg.h5"

        if os.path.exists(modelfile_cls):
            os.remove(modelfile_cls)
        if os.path.exists(modelfile_tf):
            os.remove(modelfile_tf)
        if os.path.exists(modelfile_ig_tf):
            os.remove(modelfile_ig_tf)
        if os.path.exists(modelfile_cg_tf):
            os.remove(modelfile_cg_tf)

        self.model.save(modelfile_tf)
        self.model_image_generator.save(modelfile_ig_tf)
        self.model_code_generator.save(modelfile_cg_tf)

        model = self.model
        model_image_generator = self.model_image_generator
        model_code_generator = self.model_code_generator
        metrics = self.metrics

        self.model = None
        self.model_image_generator = None
        self.model_code_generator = None
        self.metrics = None
        self.optimizer = None
        self.loss_function = None
        self.input = None
        self.code = None

        print("Pickle protocol:", pickle.HIGHEST_PROTOCOL)
        with open(modelfile_cls, "wb") as f:
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)

        self.model = model
        self.model_image_generator = model_image_generator
        self.model_code_generator = model_code_generator
        self.metrics = metrics

        # Create a ZipFile Object
        with ZipFile(modelfile, 'w') as zipobj:
           # Add multiple files to the zip
           zipobj.write(modelfile_cls)
           zipobj.write(modelfile_tf)
           zipobj.write(modelfile_ig_tf)
           zipobj.write(modelfile_cg_tf)

        os.remove(modelfile_cls)
        os.remove(modelfile_tf)
        os.remove(modelfile_ig_tf)
        os.remove(modelfile_cg_tf)

        model_dir = "model.tf"
        if os.path.exists(model_dir):
            print("Removing old model directory:", model_dir)
            shutil.rmtree(model_dir)



import os
import math
import time
import pickle
import shutil
import random
import numpy as np
from zipfile import ZipFile
import tensorflow as tf
from tensorflow import keras
import tensorflow.keras.backend as K
from tensorflow.keras import layers
from tensorflow.keras.callbacks import ModelCheckpoint
from tqdm import tqdm
from datetime import datetime
from sklearn.model_selection import train_test_split
from gimmick import constants

class Generator():
    def __init__(self, optimizer, learning_rate, num_layers=None):
        super(Generator, self).__init__()

        self.learning_rate = learning_rate
        self.optimizer = optimizer
        self.num_layers = num_layers

    def build_graph(self, images_shape, code_length, hidden_dim):
        '''
        Generator Class
        Parameters:
            code_length: the dimension of the noise vector, a scalar
            total_image_pixels: the dimension of the images, fitted for the dataset used, a scalar
              (MNIST images are 28 x 28 = 784 so that is your default)
            hidden_dim: the inner dimension, a scalar
        '''
        print("======================================= Generator =========================================")
        total_image_pixels = np.prod(images_shape)
        model = keras.Sequential(name="generator")
        model.add(layers.InputLayer(input_shape=code_length))

        num_layers = self.num_layers if self.num_layers else max( 3, int(total_image_pixels / hidden_dim) - 1)
        print('layers', num_layers)

        for i in range(num_layers - 1):
            model.add(layers.Dense(hidden_dim * i, name="generator_layer_" + str(i)))
            model.add(layers.BatchNormalization())
            model.add(layers.Activation('relu'))

        model.add(layers.Dense(total_image_pixels, name="generator_layer_last" + str(i)))
        #model.add(layers.BatchNormalization())
        model.add(layers.Activation('sigmoid'))
        model.add(layers.Reshape(images_shape))

        self.optimizer.learning_rate = self.learning_rate
        #model.compile(optimizer=self.generator_optimizer, loss=self.loss_function, metrics=self.metrics)
        print(model.summary())
        self.model = model

class Discriminator():
    def __init__(self, optimizer, learning_rate, num_layers=None):
        super(Discriminator, self).__init__()

        self.learning_rate = learning_rate
        self.optimizer = optimizer
        self.num_layers = num_layers

    def build_graph(self, images_shape, hidden_dim):
        '''
        Discriminator Class
        Parameters:
            code_length: the dimension of the noise vector, a scalar
            total_image_pixels: the dimension of the images, fitted for the dataset used, a scalar
              (MNIST images are 28 x 28 = 784 so that is your default)
            hidden_dim: the inner dimension, a scalar
        '''
        print("======================================= Discriminator =========================================")
        total_image_pixels = np.prod(images_shape)
        num_layers = self.num_layers if self.num_layers else int(math.log(total_image_pixels, 2)) - int(math.log(hidden_dim, 2))
        print('layers', num_layers)

        model = keras.Sequential(name="discriminator")
        model.add(layers.InputLayer(input_shape=images_shape))

        model.add(layers.Flatten())
        for i in range(num_layers, 0, -1):
            model.add(layers.Dense(hidden_dim * i, name="discriminator_layer_" + str(i)))
            model.add(tf.keras.layers.LeakyReLU())
            model.add(layers.Dropout(0.3))

        model.add(layers.Dense(1))
        #model.add(layers.Activation('sigmoid'))
        self.optimizer.learning_rate = self.learning_rate
        print(model.summary())
        self.model = model


class Model():

    def __init__(self, learning_rate=None, optimizer=None, optimizer_keys=None, loss_function=None, loss_function_keys=None, metrics=None, metrics_keys=None,
                 code_length=8, num_generator_layers=-1, num_discriminator_layers=-1):

        self.generator = Generator(optimizer, learning_rate, num_layers=3)
        self.discriminator = Discriminator(optimizer, learning_rate)

        self.loss_function = loss_function
        self.metrics = metrics
        self.code_length = code_length

        self.loss_function_keys = loss_function_keys
        self.optimizer_keys = optimizer_keys
        self.metrics_keys = metrics_keys

    def build_model_graph(self, images_shape):
        print("images_shape:", images_shape)

        total_image_pixels = np.prod(images_shape)
        self.generator.build_graph(images_shape, self.code_length, hidden_dim=128)
        self.discriminator.build_graph(images_shape, hidden_dim=128)

    @tf.function
    def train_step(self, images, code_length, batch_size, eval=False):
        training = not eval
        def discriminator_loss(pred_real, pred_fake):
            real_loss = self.loss_function(tf.ones_like(pred_real), pred_real) # compares discriminator's predictions on real images to an array of 1s,
            fake_loss = self.loss_function(tf.zeros_like(pred_fake), pred_fake) # compares discriminator's predictions on fake (generated) images to an array of 0s.
            total_loss = (real_loss + fake_loss) / 2
            return total_loss

        def generator_loss(pred_fake):
            # Loss will be low is fake images have a higher score by discriminator of being 1, else viseversa
            return self.loss_function(tf.ones_like(pred_fake), pred_fake) # compares discriminators decisions on the generated images to an array of 1s.

        noise = tf.random.normal([batch_size, code_length])

        with tf.GradientTape() as gen_tape, tf.GradientTape() as disc_tape:

            generated_images = self.generator.model(noise, training=training)

            pred_real = self.discriminator.model(images, training=training)
            pred_fake = self.discriminator.model(generated_images, training=training)

            gen_loss = generator_loss(pred_fake)
            disc_loss = discriminator_loss(pred_real, pred_fake)

        if training:
            gradients_of_generator = gen_tape.gradient(gen_loss, self.generator.model.trainable_variables)
            gradients_of_discriminator = disc_tape.gradient(disc_loss, self.discriminator.model.trainable_variables)

            self.generator.optimizer.apply_gradients(zip(gradients_of_generator, self.generator.model.trainable_variables))
            #if random.randint(1, 2) % 2 == 0:
            self.discriminator.optimizer.apply_gradients(zip(gradients_of_discriminator, self.discriminator.model.trainable_variables))
        return gen_loss, disc_loss

    def train(self, images_train, images_test, epochs=10, batch_size=16, validation_split=0.2):

        checkpoint = tf.train.Checkpoint(generator_optimizer=self.generator.optimizer, discriminator_optimizer=self.discriminator.optimizer,
                                         generator=self.generator.model,  discriminator=self.discriminator.model)

        # early_stopping = tf.keras.callbacks.EarlyStopping(monitor='val_loss', patience=5, restore_best_weights=True)
        print("================================= Training ===================================")

        generator = self.generator
        discriminator = self.discriminator
        code_length = self.code_length

        images_train, images_val = train_test_split(images_train, random_state=0, test_size=validation_split, shuffle=True)

        train_dataset = tf.data.Dataset.from_tensor_slices(images_train).shuffle(batch_size * 4).batch(batch_size)
        val_dataset = tf.data.Dataset.from_tensor_slices(images_val).shuffle(batch_size * 4).batch(batch_size)
        test_dataset = tf.data.Dataset.from_tensor_slices(images_test).shuffle(batch_size * 4).batch(batch_size)

        def _epoch(images, dataset, type, eval, epoch=None):
            total_generator_loss = 0
            total_discriminator_loss = 0
            steps = images.shape[0] / batch_size

            batch_log = tqdm(total=steps, position=0)

            for j, image_batch in enumerate(dataset):
                generator_loss, discriminator_loss = self.train_step(image_batch, code_length, batch_size, eval=eval)

                # Adjust weights only when training
                total_discriminator_loss += K.eval(discriminator_loss)
                total_generator_loss  += K.eval(generator_loss)

                mean_generator_loss = total_generator_loss / j
                mean_discriminator_loss = total_discriminator_loss / j

                batch_log.update(1)
                if epoch is not None:
                    batch_log.set_description(f'epoch: {i}, generator_loss({type}): {mean_generator_loss:10.8f}, discriminator_loss({type}): {mean_discriminator_loss: 10.8f}')
                else:
                    batch_log.set_description(f'generator_loss({type}): {mean_generator_loss:10.8f}, discriminator_loss({type}): {mean_discriminator_loss: 10.8f}')

        startime = datetime.now()
        for i in range(epochs):
            _epoch(images_train, train_dataset, "train", eval=False, epoch=i)
            _epoch(images_val, val_dataset, "val", eval=True, epoch=i)
            checkpoint.save(constants.DEFAULT_TF_CHECKPOINT)

        print("Total Training time:", datetime.now() - startime)

            # generator.fit(images_train, images_train, batch_size=batch_size, epochs=1, validation_split=validation_split,
            #       callbacks=[checkpoint, early_stopping], shuffle=True)
        #model.save(constants.DEFAULT_TF_MODELFILE) # Save Best model to disk

        print("================================= Evaluating ===================================")
        _epoch(images_val, val_dataset, "test", eval=True)

    def generate(self, n, codes=None, batch_size=8, print_model=False):
        """ This function is used to train model on given training images, based on data enoder part learns to compress images into intermediate dimension containing N points, and decoder learns to decipher the intermediate dimension. The indermediate dimention of N points (code_length), when randomly generated will produce new images.

        **Parameters**

        n: int
            number of images to be generated by our trained model
        codes: list
            Default: None, if passes then model will use this set of codes for generating images, else it will generate random distribution based on earlier calculated code statistics
        batch_size: int
            Default 16, batch_size used for training model.
        print_model: boolean
            Default False,  whether to print model information
        """
        print("================================= generating samples ===================================")
        if print_model:
            print(self.generator.model.summary())

        if codes is not None:
            inputs  = codes[:n]
        else:
            inputs = tf.random.normal([n, self.code_length])

        dataset = tf.data.Dataset.from_tensor_slices(inputs).shuffle(batch_size * 4).batch(batch_size)

        images_generated = []
        for i, noise_batch in enumerate(dataset):
            images_generated.append(self.generator.model(noise_batch, training=True))
        print(images_generated[0][0])
        images_generated = np.concatenate(images_generated)#.astype(np.uint8)
        #images_generated[images_generated > 255] = 255
        #images_generated[images_generated < 0] = 0
        return images_generated

    def save(self, modelfile):
        """ Function save model to a persistat storage

        **Parameters**

        modelfile: string
            Path where model needs to be saved
        """
        if not modelfile.endswith('.zip'):
            raise Exception('modelfile must ends with .zip as extention')

        if os.path.exists(modelfile):
            os.remove(modelfile)

        modelfile_cls = "tf_" + modelfile.split('.')[0] + ".pkl"
        modelfile_gen_tf = "tf_" + modelfile.split('.')[0] + "_gen.h5"
        modelfile_disc_tf = "tf_" + modelfile.split('.')[0] + "_disc.h5"

        self.generator.model.save(modelfile_gen_tf)
        self.discriminator.model.save(modelfile_disc_tf)

        generator = self.generator
        discriminator = self.discriminator

        metrics = self.metrics
        loss_function = self.loss_function

        self.discriminator.model = None
        self.discriminator.optimizer = None
        self.generator.model = None
        self.generator.optimizer = None
        self.metrics = None
        self.loss_function = None

        print("Pickle protocol:", pickle.HIGHEST_PROTOCOL)
        with open(modelfile_cls, "wb") as f:
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)

        self.generator = generator
        self.discriminator = discriminator
        self.metrics = metrics
        self.loss_function = loss_function

        with ZipFile(modelfile, 'w') as zipobj:
           zipobj.write(modelfile_cls)
           zipobj.write(modelfile_gen_tf)
           zipobj.write(modelfile_disc_tf)

        os.remove(modelfile_cls)
        os.remove(modelfile_gen_tf)
        os.remove(modelfile_disc_tf)

        model_dir = constants.DEFAULT_TF_CHECKPOINT_DIR
        if os.path.exists(model_dir):
            print("Removing old model directory:", model_dir)
            shutil.rmtree(model_dir)



""" This module provide common utility function on images which are used across different sections
"""
import os
import cv2
import math
import numpy as np
from PIL import Image
from matplotlib import pyplot as plt
from mpl_toolkits.axes_grid1 import ImageGrid

def convert_2dto3d(images):
    """ Function convert a list of 2D images to list of 3D images
    """

    print("convering images to 3d: shape before convert_2dto3d", images.shape)
    if len(images.shape) == 3:
        images = images.reshape(images.shape[0], images.shape[1], images.shape[2], 1)
    print("shape after convert_2dto3d", images.shape)
    return images

def plot_images(images, n_col=8, outputfile_path='img.png'):
    n_row = int(math.ceil(len(images) / n_col))
    fig = plt.figure(figsize=(6., 6.))

    grid = ImageGrid(fig, 111,  # similar to subplot(111)
                 nrows_ncols=(n_row, n_col),  # creates 2x2 grid of axes
                 axes_pad=0.0,  # pad between axes in inch.
    )

    import cv2
    for ax, img in zip(grid, images):
        # Iterating over the grid returns the Axes.
        if type(img) == str and os.path.exists(img):
            img = cv2.imread(img)
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
#             img = cv2.resize(img, (64, 64))  # Reshaping for visualization

        if len(img.shape) == 3 and img.shape[2] == 1:
            img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
        ax.imshow(img)
    plt.show()

    colored_image  = outputfile_path

    gray_image  = 'gray_' + outputfile_path
    plt.savefig(colored_image)
    Image.open(colored_image).convert('L').save(gray_image)
    return colored_image, gray_image

def read_images_from_disk(trainig_files):
    images = [cv2.imread(x) for x in trainig_files]
    images = np.array(images)
    return images

def rescale_images(images, img_width, img_height, img_channels):
    images = [cv2.resize(x, (img_width, img_height)) for x in images]  # Reshaping for visualization
    images = np.array(images)
    return images

if __name__ == "__main__":
    from sklearn import datasets
    digits = datasets.load_digits()
    images = digits.images  # It contains roughly 1800 images of shape 8 x 8
    print("shape before convert_2dto3d", images.shape)
    images = convert_2dto3d(images)
    print("shape after convert_2dto3d", images.shape)



import os
import pickle
import tensorflow as tf
from gimmick.distributer import *
from gimmick import mapping
from zipfile import ZipFile

def load(modelfile):
    modelfile_cls = "tf_" + modelfile.split('.')[0] + ".pkl"

    with ZipFile(modelfile, 'r') as zip:
        zip.printdir()
        zip.extractall()

    modelobj = pickle.load(open(modelfile_cls, 'rb'))

    def _load_gans():
        modelfile_tf = "tf_" + modelfile.split('.')[0] + ".h5"
        modelfile_gen_tf = "tf_" + modelfile.split('.')[0] + "_gen.h5"
        modelfile_disc_tf = "tf_" + modelfile.split('.')[0] + "_disc.h5"

        modelobj.generator.model = tf.keras.models.load_model(modelfile_gen_tf)
        modelobj.discriminator.model = tf.keras.models.load_model(modelfile_disc_tf)

        os.remove(modelfile_cls)
        os.remove(modelfile_gen_tf)

        modelobj.metrics = [mapping.metrics_mapping.get(x) for x in modelobj.metrics_keys]
        modelobj.loss_function = mapping.loss_function_mapping.get(modelobj.loss_function_keys)
        modelobj.generator.optimizer = mapping.optimizer_mapping.get(modelobj.optimizer_keys)
        modelobj.discriminator.optimizer = mapping.optimizer_mapping.get(modelobj.optimizer_keys)

        return modelobj

    def _load_autoencoders():
        modelfile_tf = "tf_" + modelfile.split('.')[0] + ".h5"
        modelfile_ig_tf = "tf_" + modelfile.split('.')[0] + "_ig.h5"
        modelfile_cg_tf = "tf_" + modelfile.split('.')[0] + "_cg.h5"

        modelobj.model = tf.keras.models.load_model(modelfile_tf)
        modelobj.model_image_generator = tf.keras.models.load_model(modelfile_ig_tf)
        modelobj.model_code_generator = tf.keras.models.load_model(modelfile_cg_tf)
        os.remove(modelfile_ig_tf)
        os.remove(modelfile_cg_tf)
        os.remove(modelfile_cls)
        os.remove(modelfile_tf)

        modelobj.metrics = [mapping.metrics_mapping.get(x) for x in modelobj.metrics_keys]
        modelobj.optimizer = mapping.optimizer_mapping.get(modelobj.optimizer_keys)
        modelobj.loss_function = mapping.loss_function_mapping.get(modelobj.loss_function_keys)
        return modelobj

    if ".gan" in str(modelobj):
        return _load_gans()
    return _load_autoencoders()



import math
import pickle
import numpy as np
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers
from gimmick import constants
from gimmick.models.autoencoder import AutoEncoder

class Model(AutoEncoder):
    """ This algorithm uses a CNN based variational autoenocder for learning and generating images.
    https://en.wikipedia.org/wiki/Variational_autoencoder
    """

    def build_model_graph(self, images_shape):
        print(self.__dict__)
        total_pixels = images_shape[0] * images_shape[1] * images_shape[2]

        num_encoder_layers = int(math.log(images_shape[0], 2)) - 2 #if self.num_encoder_layer == "auto" else int(self.num_encoder_layer)
        num_decoder_layers = int(math.log(images_shape[0], 2)) - 2 #if self.num_decoder_layers == "auto" else int(self.num_decoder_layers)

        log2_code = int(math.log(self.code_length, 2))
        print("num_enoder_layer:\t", num_encoder_layers)
        print("num_decoder_layers:\t", num_decoder_layers)
        print("log2_code:\t\t", log2_code)

        filter_size = (3, 3)
        pool_size = (2, 2)

        def _encoder(x):
            # x = layers.Reshape((images_shape[0], images_shape[1], images_shape[2]))(x)

            # Encoder Layer
            for i in range(1, num_encoder_layers + 1):
                neurons = 2 ** (num_encoder_layers - i + log2_code + 1) # Encoder layer size will be always greater then code_length by multiple of 2

                if i == 1 and images_shape[0] <= 16:
                    x = layers.Conv2D(neurons, filter_size, activation='relu', padding='same', strides=1, name="encoder_layer_extra1")(x)
                if i == 1 and images_shape[0] <= 8:
                    x = layers.Conv2D(neurons, filter_size, activation='relu', padding='same', strides=1, name="encoder_layer_extra2")(x)
                x = layers.Conv2D(neurons, filter_size, activation='relu', padding='same', strides=2, name="encoder_layer_" + str(i) )(x)

            # Code Layer
            x = layers.Flatten()(x)
            return x

        def _decoder(code):

            x = layers.Reshape((2, 2, int(self.code_length / 4)))(code)

            # Decoder Layer
            for i in range(1, num_decoder_layers + 1):
                neurons = 2 ** (i + log2_code)  # Decoder layer size will be always greater then code_length by multiple of 2

                if i == 1 and images_shape[0] <= 16:
                    x = layers.Conv2DTranspose(neurons, filter_size, activation='relu', padding='same', strides=1, name="decoder_layer_extra1" )(x)
                if i == 1 and images_shape[0] <= 8:
                    x = layers.Conv2DTranspose(neurons, filter_size, activation='relu', padding='same', strides=1, name="decoder_layer_extra2" )(x)
                x = layers.Conv2DTranspose(neurons, filter_size, activation='relu', padding='same', strides=2, name="decoder_layer_" + str(i) )(x)

            output = layers.Conv2DTranspose(images_shape[2], filter_size, activation='relu', padding='same', strides=2, name='decoder_layer_%s' % str(i + 1) )(x)
            return output

        input = keras.Input(shape=images_shape)

        x = _encoder(input)

        mean = layers.Dense(self.code_length, name='mean')(x)
        std = keras.activations.softplus(layers.Dense(self.code_length, name='std')(x))

        # Reparametrization trick
        epsilon = tf.random.normal(tf.stack([tf.shape(x)[0], self.code_length]), name='epsilon')
        code = tf.add(mean, epsilon * std, name='code')

        output = _decoder(code)

        optimizer =self.optimizer
        optimizer.learning_rate = self.learning_rate

        model = tf.keras.Model(inputs=input, outputs=output)
        model.compile(optimizer=optimizer, loss=self.loss_function, metrics=self.metrics)
        self.model = model

        print(model.summary())
        print("============================================================")

        def _code_generator_model():
            # Building model which generate code statistics
            model_code_generator = tf.keras.Model(inputs=input, outputs=code)
            model_code_generator.build((None, images_shape[0], images_shape[1], images_shape[2]))
            print(model_code_generator.summary())
            return model_code_generator

        self.model_code_generator = _code_generator_model()

        print("============================================================")
        def _image_generator_model():
            generator_layer_num = 0
            for cntr, layer in enumerate(model.layers):
                if layer.name == 'tf_op_layer_code':
                    break
                generator_layer_num += 1

            # Building model which generate images
            model_image_generator = keras.Sequential(model.layers[generator_layer_num+1:])
            model_image_generator.build((None, self.code_length))
            print(model_image_generator.summary())
            return model_image_generator

        self.model_image_generator = _image_generator_model()
